
import asyncio
import uuid
import networkx
import threading
import os
import time
from yaspin.spinners import Spinners
from typing import Dict, List, Any
from hedra.core.graphs.stages.base.stage import Stage
from hedra.core.graphs.stages.types.stage_types import StageTypes
from hedra.core.graphs.transitions.transition import Transition
from hedra.logging import HedraLogger
from .transitions import TransitionAssembler, local_transitions
from .status import GraphStatus



class Graph:
    status = GraphStatus.IDLE

    def __init__(
        self, 
        graph_name: str,
        stages: List[Stage], 
        config: Dict[str, Any]={},
        cpus: int=None, 
        worker_id: int=None
    ) -> None:
        self.execution_time = 0
        self.core_config = config

        self.graph_name = graph_name
        self.graph_path = config.get('graph_path')
        self.graph_id = str(uuid.uuid4())
        self.status = GraphStatus.INITIALIZING
        self.graph = networkx.DiGraph()
        self.logger = HedraLogger()
        self._thread_id = threading.current_thread().ident
        self._process_id = os.getpid()
        self.metadata_string = f'Graph - {self.graph_name}:{self.graph_id} - thread:{self._thread_id} - process:{self._process_id} - '

        self.logger.initialize()

        self.logger.hedra.sync.debug(f'{self.metadata_string} - Changed status to - {GraphStatus.INITIALIZING.name} - from - {GraphStatus.IDLE.name}')
        self.logger.filesystem.sync['hedra.core'].info(f'{self.metadata_string} - Changed status to - {GraphStatus.INITIALIZING.name} - {GraphStatus.IDLE.name}')

        self.transitions_graph = []
        self._transitions: List[List[Transition]] = []
        self._results = None

        self.logger.hedra.sync.debug(f'{self.metadata_string} - Found - {len(stages)} - stages')
        self.logger.filesystem.sync['hedra.core'].debug(f'{self.metadata_string} - Found - {len(stages)} - stages')

        self.stage_types = {
            subclass.stage_type: subclass for subclass in Stage.__subclasses__()
        }

        self.instances: Dict[StageTypes, List[Stage]] = {}
        for stage in self.stage_types.values():
            stage_instances = [stage for stage in stage.__subclasses__() if stage.__module__ == self.core_config.get('graph_module')]
            self.instances[stage.stage_type] = stage_instances

        self.stages: Dict[str, Stage] = {stage.__name__: stage for stage in stages}
        
        self.graph.add_nodes_from([
            (
                stage_name, 
                {"stage": stage}
            ) for stage_name, stage in self.stages.items()
        ])

        for stage in stages:
            
            self.logger.hedra.sync.debug(f'{self.metadata_string} - Adding dependencies for stage - {stage.__name__}')
            self.logger.filesystem.sync['hedra.core'].debug(f'{self.metadata_string} - Adding dependencies for stage - {stage.__name__}')

            for dependency in stage.dependencies:
                if self.graph.nodes.get(dependency.__name__):
                    
                    self.logger.hedra.sync.debug(f'{self.metadata_string} - Adding edge from stage - {dependency.__name__} - to stage - {stage.__name__}')
                    self.logger.filesystem.sync['hedra.core'].debug(f'{self.metadata_string} - Adding edge from stage - {dependency.__name__} - to stage - {stage.__name__}.')

                    self.graph.add_edge(dependency.__name__, stage.__name__)

        self.execution_order = [
            generation for generation in networkx.topological_generations(self.graph)
        ]

        self.runner = TransitionAssembler(
            local_transitions,
            graph_name=self.graph_name,
            graph_path=self.graph_path,
            graph_id=self.graph_id,
            cpus=cpus,
            worker_id=worker_id,
            core_config=self.core_config
        )

    def assemble(self):

        self.status = GraphStatus.ASSEMBLING

        self.logger.hedra.sync.debug(f'{self.metadata_string} - Changed status to - {GraphStatus.ASSEMBLING.name} - from - {GraphStatus.INITIALIZING.name}')
        self.logger.filesystem.sync['hedra.core'].info(f'{self.metadata_string} - Changed status to - {GraphStatus.ASSEMBLING.name} - from - {GraphStatus.INITIALIZING.name}')

        # If we havent specified a Validate stage for save aggregated results,
        # append one.

        if len(self.instances.get(StageTypes.VALIDATE)) < 1:
            self.logger.hedra.sync.debug(f'{self.metadata_string} - Prepending {StageTypes.VALIDATE.name} stage')
            self.logger.filesystem.sync['hedra.core'].debug(f'{self.metadata_string} - Prepending {StageTypes.VALIDATE.name} stage')

            self._prepend_stage(StageTypes.VALIDATE)

        # A user will never specify an Idle stage. Instead, we prepend one to 
        # serve as the source node for a graphs, ensuring graphs have a 
        # valid starting point and preventing the user from forgetting things
        # like a Setup stage.

        self.logger.hedra.sync.debug(f'{self.metadata_string} - Prepending {StageTypes.IDLE.name} stage')
        self.logger.filesystem.sync['hedra.core'].debug(f'{self.metadata_string} - Prepending {StageTypes.IDLE.name} stage')

        self._prepend_stage(StageTypes.IDLE)

        # If we haven't specified an Analyze stage for results aggregation,
        # append one.
        if len(self.instances.get(StageTypes.ANALYZE)) < 1:
            self.logger.hedra.sync.debug(f'{self.metadata_string} - Appending {StageTypes.ANALYZE.name} stage')
            self.logger.filesystem.sync['hedra.core'].debug(f'{self.metadata_string} - Appending {StageTypes.ANALYZE.name} stage')

            self._append_stage(StageTypes.ANALYZE)

        # If we havent specified a Submit stage for save aggregated results,
        # append one.
        if len(self.instances.get(StageTypes.SUBMIT)) < 1:
            self.logger.hedra.sync.debug(f'{self.metadata_string} - Appending {StageTypes.SUBMIT.name} stage')
            self.logger.filesystem.sync['hedra.core'].debug(f'{self.metadata_string} - Appending {StageTypes.SUBMIT.name} stage')

            self._append_stage(StageTypes.SUBMIT)

        # Like Idle, a user will never specify a Complete stage. We append
        # one to serve as the sink node, ensuring all Graphs executed can
        # reach a single exit point.
        self.logger.hedra.sync.debug(f'{self.metadata_string} - Appending {StageTypes.COMPLETE.name} stage')
        self.logger.filesystem.sync['hedra.core'].debug(f'{self.metadata_string} - Appending {StageTypes.COMPLETE.name} stage')

        self._append_stage(StageTypes.COMPLETE)

        self.logger.hedra.sync.debug(f'{self.metadata_string} - Generating graph stages and transitions')
        self.logger.filesystem.sync['hedra.core'].debug(f'{self.metadata_string} - Generating graph stages and transitions')

        self.runner.generate_stages(self.stages)
        self._transitions = self.runner.build_transitions_graph(self.execution_order, self.graph)
        self.runner.map_to_setup_stages(self.graph)

        self.logger.hedra.sync.debug(f'{self.metadata_string} - Assembly complete')
        self.logger.filesystem.sync['hedra.core'].debug(f'{self.metadata_string} - Assembly complete')

    async def run(self):

        execution_start = time.monotonic()

        await self.logger.hedra.aio.debug(f'{self.metadata_string} - Changed status to - {GraphStatus.RUNNING.name} - from - {GraphStatus.ASSEMBLING.name}')
        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Changed status to - {GraphStatus.RUNNING.name} - from - {GraphStatus.ASSEMBLING.name}')
        
        self.status = GraphStatus.RUNNING

        for transition_group in self._transitions:

            current_stages = ', '.join([transition.from_stage.name for transition in transition_group])

            async with self.logger.spinner as status_spinner:
                
                await self.logger.spinner.append_message(f"Executing stages - {current_stages}")
                
                for transition in transition_group:

                    await status_spinner.system.debug(f'{self.metadata_string} - Executing stage Transtition - {transition.transition_id} -  from stage - {transition.from_stage.name} - to stage - {transition.to_stage.name}')
                    await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Executing stage Transition - {transition.transition_id} - from stage - {transition.from_stage.name} - to stage - {transition.to_stage.name}')
                    await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Executing stage - {transition.from_stage.name}:{transition.from_stage.stage_id}')

                results = await asyncio.gather(*[
                    asyncio.create_task(transition.transition(
                        transition.from_stage, 
                        transition.to_stage
                    )) for transition in transition_group
                ])

                for transition in transition_group:
                    await status_spinner.system.debug(f'{self.metadata_string} - Completed stage Transtition - {transition.transition_id} -  from stage - {transition.from_stage.name} - to stage - {transition.to_stage.name}')

                completed_transitions_count = len(results)
                await status_spinner.system.debug(f'{self.metadata_string} - Completed -  {completed_transitions_count} - transitions')
                await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Completed -  {completed_transitions_count} - transitions')

                for error, next_stage in results:
                    
                    if next_stage == StageTypes.ERROR:

                        self.status = GraphStatus.FAILED

                        await status_spinner.system.debug(f'{self.metadata_string} - Changed status to - {GraphStatus.FAILED.name} - from - {GraphStatus.RUNNING.name}')
                        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Changed status to - {GraphStatus.FAILED.name} - from - {GraphStatus.RUNNING.name}')

                        await status_spinner.system.error(f'{self.metadata_string} - Encountered error executing stage - {error.from_stage.name}:{error.from_stage.stage_id}')
                        await self.logger.filesystem.aio['hedra.core'].error(f'{self.metadata_string} - Encountered error executing stage - {error.from_stage.name}:{error.from_stage.stage_id}')

                        error_transtiton = self.runner.create_error_transition(error)

                        await error_transtiton.transition(
                            error_transtiton.from_stage,
                            error_transtiton.to_stage
                        ) 
                    
                if self.status == GraphStatus.FAILED:
                    status_spinner.finalize()
                    await status_spinner.fail('Error')
                    break

                status_spinner.group_finalize()

                await status_spinner.ok('✔')
            
        if self.status == GraphStatus.RUNNING:
            await status_spinner.system.debug(f'{self.metadata_string} - Changed status to - {GraphStatus.COMPLETE.name} - from - {GraphStatus.RUNNING.name}')
            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Changed status to - {GraphStatus.COMPLETE.name} - from - {GraphStatus.RUNNING.name}')

            self.status = GraphStatus.COMPLETE

        self.execution_time = execution_start - time.monotonic()

        return self._results

    async def check(self, graph_path: str):
        
        validation_stages = []

        for transition_group in self._transitions:
            for transtition in transition_group:

                if transtition.from_stage.stage_type == StageTypes.VALIDATE: 
                    validation_stages.append(transtition)

        await asyncio.gather(*[
            asyncio.create_task(transition.transition(
                transition.from_stage, 
                transition.to_stage
            )) for transition in validation_stages
        ])

    async def cleanup(self):
        pass

    def _append_stage(self, stage_type: StageTypes):

        appended_stage = self.stage_types.get(stage_type)
        last_cut = self.execution_order[-1]

        appended_stage.dependencies = list()

        for stage_name in last_cut:
            stage = self.stages.get(stage_name)

            appended_stage.dependencies.append(stage)

        self.graph.add_node(appended_stage.__name__, stage=appended_stage)

        for stage_name in last_cut:
            self.graph.add_edge(stage_name, appended_stage.__name__)

        self.execution_order = [
            generation for generation in networkx.topological_generations(self.graph)
        ]

        self.stages[appended_stage.__name__] = appended_stage
        self.instances[appended_stage.stage_type].append(appended_stage)

    def _prepend_stage(self, stage_type: StageTypes):
        prepended_stage = self.stage_types.get(stage_type)
        first_cut = self.execution_order[0]

        prepended_stage.dependencies = list()

        for stage_name in first_cut:
            stage = self.stages.get(stage_name)

            stage.dependencies.append(prepended_stage)

        self.graph.add_node(prepended_stage.__name__, stage=prepended_stage)

        for stage_name in first_cut:
            self.graph.add_edge(prepended_stage.__name__, stage_name)

        self.execution_order = [
            generation for generation in networkx.topological_generations(self.graph)
        ]

        self.stages[prepended_stage.__name__] = prepended_stage
        self.instances[prepended_stage.stage_type].append(stage_type)



from .registered_worker import RegisteredWorker


class WorkerRegistry:

    def __init__(self, worker_addresses, workers_count=None) -> None:
        logger = Logger()
        self.session_logger = logger.generate_logger()
        self.worker_addresses = worker_addresses
        self.registered_workers = {}
        self.failed = []
        self.initial_count = workers_count if workers_count else len(worker_addresses)
        self.count = 0

    def __getitem__(self, worker_address):
        return self.registered_workers[worker_address]

    def __iter__(self):
        for worker in dict(self.registered_workers):
            yield self.registered_workers[worker]

    async def __aiter__(self):
        for worker in self.registered_workers:
            yield self.registered_workers[worker]

    async def initialize(self):
        for worker_address in self.worker_addresses:
            await self.register_worker(worker_address)

    async def register_worker(self, worker_address):
        if worker_address not in self.registered_workers:
            self.session_logger.info(f'Registering new worker from - {worker_address}.')
            self.registered_workers[worker_address] = RegisteredWorker(worker_address)
            self.count += 1

        if worker_address in self.failed:
            self.session_logger.info(f'Worker at - {worker_address} - re-registered.')
            self.failed.remove(worker_address)

        if worker_address not in self.worker_addresses:
            self.worker_addresses.append(worker_address)

    async def worker_id_not_set(self, worker_address):
        return self.registered_workers[worker_address].worker_id is None

    async def set_worker_id(self, worker_address, worker_id):
        self.registered_workers[worker_address].worker_id = worker_id

    async def remove_worker(self, worker_address):
        if worker_address in self.registered_workers and worker_address in self.worker_addresses:
            self.failed.append(worker_address)
            self.worker_addresses.remove(worker_address)
            self.count -= 1

    async def clear_registry(self):
        self.worker_addresses = []
        self.registered_workers = {}
        self.count = 0

    async def clear_failed(self):
        self.failed = []
        
    async def registered_count_equals_expected_count(self):
        return len(self.registered_workers) == self.initial_count


import grpc
from hedra.runners.worker_services.proto import (
    WorkerServerStub
)

class RegisteredWorker:

    def __init__(self, worker_address):
        self.worker_address = worker_address
        self.worker_id = None

        channel = grpc.aio.insecure_channel(worker_address)
        service = WorkerServerStub(channel)

        self.client = service



import os
from hedra.core.graphs.stages import (
    Submit
)

from hedra.reporting.types import (
    MySQLConfig
)


class SubmitMySQLResultsStage(Submit):
    config=MySQLConfig(
        host='127.0.0,1',
        database='results',
        username=os.getenv('MYSQL_USERNAME', ''),
        password=os.getenv('MYSQL_PASSWORD', ''),
        events_table='events',
        metrics_table='metrics'
    )


# -*- coding: utf-8 -*-
"""
h2/frame_buffer
~~~~~~~~~~~~~~~

A data structure that provides a way to iterate over a byte buffer in terms of
frames.
"""
import h2.errors
import sys
import struct
from .types.attributes import (
    _STRUCT_HBBBL
)
from .types import *
from hyperframe.exceptions import InvalidFrameError, InvalidDataError
from .types.base_frame import Frame
from .types.attributes import (
    Flag,
    Flags,
    _STRUCT_HBBBL,
    _STRUCT_H,
    _STRUCT_B,
    _STRUCT_LL,
    _STRUCT_LB,
    _STRUCT_L,
    _STRUCT_HL
)


class FrameBuffer:

    __slots__ = (
        'data',
        'max_frame_size',
        '_headers_buffer',
        'frames_map'
    )

    """
    This is a data structure that expects to act as a buffer for HTTP/2 data
    that allows iteraton in terms of H2 frames.
    """
    def __init__(self):
        self.data = bytearray()
        self.max_frame_size = 0
        self._headers_buffer = []

    # The methods below support the iterator protocol.
    def __iter__(self):
        while len(self.data) >= 9:
            try:
                
                fields = _STRUCT_HBBBL.unpack(self.data[:9])

                # First 24 bits are frame length.
                length = (fields[0] << 8) + fields[1]
                type = fields[2]
                flags = fields[3]
                stream_id = fields[4] & 0x7FFFFFFF

                frame = Frame(stream_id, type, parsed_flag_byte=flags)

            except (InvalidDataError, InvalidFrameError) as e:  # pragma: no cover
                raise Exception(
                    "Received frame with invalid header: %s" % str(e)
                )

            # Next, check that we have enough length to parse the frame body. If
            # not, bail, leaving the frame header data in the buffer for next time.
            if len(self.data) < length + 9:
                break

            body_data = self.data[9:9+length]

            if frame.type == 0xA:
                # ALTSVC

                origin_len = _STRUCT_H.unpack(body_data[0:2])[0]
                frame.origin = body_data[2:2+origin_len]

                if len(frame.origin) != origin_len:
                    raise Exception("Invalid ALTSVC frame body.")

                frame.field = body_data[2+origin_len:]
                frame.body_len = len(body_data)

            elif frame.type == 0x09:
                # CONTINUATION
                frame.data = body_data
                frame.body_len = len(body_data)

            elif frame.type == 0x0:
                # DATA
                padding_data_offset = 0
                frame.pad_length = 0

                if 'PADDED' in frame.flags:  # type: ignore
                    frame.pad_length = struct.unpack('!B', body_data[:1])[0]
                    padding_data_offset = 1

                data_length = len(body_data)
                frame.data = (
                    body_data[padding_data_offset:data_length-frame.pad_length]
                )
                frame.body_len = data_length

            elif frame.type == 0x07:
                # GOAWAY

                frame.last_stream_id, frame.error_code = _STRUCT_LL.unpack(body_data[:8])
                frame.body_len = len(body_data)

                if len(body_data) > 8:
                    frame.additional_data = body_data[8:]

            elif frame.type == 0x01:
                # HEADERS
                padding_data_offset = 0

                if 'PADDED' in frame.flags:  # type: ignore
                    frame.pad_length = struct.unpack('!B', body_data[:1])[0]
                    padding_data_offset = 1
        
                body_data = body_data[padding_data_offset:]

                if 'PRIORITY' in frame.flags:
                    frame.depends_on, frame.stream_weight = _STRUCT_LB.unpack(body_data[:5])
                    frame.exclusive = True if frame.depends_on >> 31 else False
                    frame.depends_on &= 0x7FFFFFFF
                    padding_data_offset = 5

                else:
                    padding_data_offset = 0

                data_length = len(body_data)
                frame.body_len = data_length
                frame.data = body_data[padding_data_offset:data_length-frame.pad_length]

            elif frame.type == 0x06:
                # PING
                frame.opaque_data = body_data
                frame.body_len = 8

            elif frame.type == 0x02:
                # PRIORITY

                try:
                    frame.depends_on, frame.stream_weight = _STRUCT_LB.unpack(body_data[:5])
                except struct.error:
                    raise Exception("Invalid Priority data")

                frame.exclusive = True if frame.depends_on >> 31 else False
                frame.depends_on &= 0x7FFFFFFF

                frame.body_len = 5

            elif frame.type == 0x05:
                # PUSH PROMISE
                padding_data_offset = 0
                frame.pad_length = 0
                if 'PADDED' in frame.flags:  # type: ignore
                    try:
                        frame.pad_length = struct.unpack('!B', body_data[:1])[0]
                    except struct.error:
                        raise Exception("Invalid Padding data")
                    padding_data_offset = 1

                frame.promised_stream_id = _STRUCT_L.unpack(
                    body_data[padding_data_offset:padding_data_offset + 4]
                )[0]

                data_len = len(body_data)
                frame.data = body_data[padding_data_offset + 4:data_len-frame.pad_length]
                frame.body_len = data_len

            elif frame.type == 0x03:
                # RESET
                frame.error_code = _STRUCT_L.unpack(body_data)[0]
                frame.body_len = 4

            elif frame.type == 0x04:
                # SETTINGS
                body_len = 0
                for i in range(0, len(body_data), 6):

                    name, value = _STRUCT_HL.unpack(body_data[i:i+6])

                    frame.settings[name] = value
                    body_len += 6

                frame.body_len = body_len

            elif frame.type == 0x08:
                # WINDOW UPDATE
                frame.window_increment = _STRUCT_L.unpack(body_data)[0]
                frame.body_len = 4

            # At this point, as we know we'll use or discard the entire frame, we
            # can update the data.
            del self.data[:9+length]

            # Pass the frame through the heaer buffer.
            # f = self._update_header_buffer(f)
            is_headers_or_push_promise = frame.frame_type == 'HEADERS' or frame.frame_type == 'PUSHPROMISE'

            if self._headers_buffer:
                stream_id = self._headers_buffer[0].stream_id
                valid_frame = (
                    frame is not None and
                    frame.frame_type == 'CONTINUATION' and
                    frame.stream_id == stream_id
                )

                if valid_frame == False:
                    raise Exception("Invalid frame during header block.")

                # Append the frame to the buffer.
                self._headers_buffer.append(frame)

                # If this is the end of the header block, then we want to build a
                # mutant HEADERS frame that's massive. Use the original one we got,
                # then set END_HEADERS and set its data appopriately. If it's not
                # the end of the block, lose the current frame: we can't yield it.
                if 'END_HEADERS' in frame.flags:
                    frame = self._headers_buffer[0]
                    frame.flags.add('END_HEADERS')

                    frame_data = bytearray()
                    for header_frame in self._headers_buffer:
                        frame_data.extend(header_frame)

                    frame.data = frame_data
                    self._headers_buffer = []

                else:
                    frame = None
            elif is_headers_or_push_promise and 'END_HEADERS' not in frame.flags:
                # This is the start of a headers block! Save the frame off and then
                # act like we didn't receive one.
                self._headers_buffer.append(frame)
                frame = None

            # If we got a frame we didn't understand or shouldn't yield, rather
            # than return None it'd be better if we just tried to get the next
            # frame in the sequence instead. Recurse back into ourselves to do
            # that. This is safe because the amount of work we have to do here is
            # strictly bounded by the length of the buffer.
            if frame:
                yield frame



from collections import defaultdict
from typing import Dict, List
from hedra.core.graphs.hooks.hook_types.hook_type import HookType
from hedra.core.graphs.hooks.registry.registry_types import (
    ActionHook,
    TaskHook,
    ValidateHook
)
from hedra.core.graphs.hooks.registry.registry_types.hook import Hook
from hedra.core.graphs.stages.base.stage import Stage
from hedra.core.graphs.stages.types.stage_types import StageTypes
from hedra.logging import HedraLogger
from .hook_validators import (
    ActionHookValidator,
    AfterHookValidator,
    BaseHookVaidator,
    BeforeHookVaidator,
    ChannelHookValidator,
    CheckHookValidator,
    EventHookVaidator,
    MetricHookVaidator,
    RestoreHookValidator,
    SaveHookValidator,
    SetupHookVaidator,
    TaskHookValidator,
    TeardownHookValidator,
    ValidateHookValidator
)
from .stage_validator import StageValidator



class Validator:

    def __init__(self, stages: Dict[str, Dict[str, Stage]], metadata_string: str) -> None:


        self.logger = HedraLogger()
        self.logger.initialize()
        self.metadata_string: str = metadata_string

        self.hooks_by_name: Dict[str, Hook] = defaultdict(dict)
        self.custom_validation_hooks: List[ValidateHook] = []

        self._validators = {
            HookType.ACTION: ActionHookValidator(metadata_string),
            HookType.AFTER: AfterHookValidator(metadata_string),
            HookType.BEFORE: BeforeHookVaidator(metadata_string),
            HookType.CHANNEL: ChannelHookValidator(metadata_string),
            HookType.CHANNEL: CheckHookValidator(metadata_string),
            HookType.EVENT: EventHookVaidator(metadata_string),
            HookType.METRIC: MetricHookVaidator(metadata_string),
            HookType.RESTORE: RestoreHookValidator(metadata_string),
            HookType.SAVE: SaveHookValidator(metadata_string),
            HookType.SETUP: SetupHookVaidator(metadata_string),
            HookType.TASK: TaskHookValidator(metadata_string),
            HookType.TEARDOWN: TeardownHookValidator(metadata_string),
            HookType.VALIDATE: ValidateHookValidator(metadata_string)
        }

        self.validation_stages = {
            stage_type: stage for stage_type, stage in stages.items() if stage_type not in [
                StageTypes.VALIDATE, 
                StageTypes.IDLE, 
                StageTypes.COMPLETE
            ]
        }

        self.stage_names: List[str] = []
        for stage_type in self.validation_stages:
            self.stage_names.extend(self.validation_stages[stage_type])

        self.hooks = {
            hook_type: [] for hook_type in HookType
        }


        self.stage_validators: List[StageValidator] = []

        for stage_type in self.validation_stages:
            for stage_name in self.validation_stages[stage_type]:
                stage = self.validation_stages[stage_type].get(stage_name)
                self.stage_validators.append(StageValidator(stage, metadata_string))

    def add_hooks(self, hook_type: HookType, hooks: List[Hook]):
        self.hooks[hook_type].extend(hooks)

    async def validate_stages(self):

        validation_stage_names = ', '.join(self.stage_names)

        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Validating stages - {validation_stage_names}')
        await self.logger.spinner.append_message(f'Validating - {len(self.validation_stages)} - stages')
        
        for stage_validator in self.stage_validators:
            await stage_validator.validate_stage_internal_hooks()
            await stage_validator.validate_stage_hooks()

            for hook_type, hooks in stage_validator.hooks.items():
                self.hooks[hook_type].extend(hooks)

            self.hooks_by_name.update(stage_validator.hooks_by_name)

        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Validation for stages - {validation_stage_names} - complete')
        await self.logger.spinner.set_default_message(f'Validated - {len(self.stage_names)} stages')

    async def validate_hooks(self):

        channel_hook_validator: ChannelHookValidator = self._validators[HookType.CHANNEL]
        channel_hook_validator.action_and_task_hooks = [
            *self.hooks.get(HookType.ACTION),
            *self.hooks.get(HookType.TASK)
        ]

        self._validators[HookType.CHANNEL] = channel_hook_validator

        for hook_type in self.hooks:
            
            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Validating - {hook_type.name.capitalize()} - hooks')

            validator: BaseHookVaidator = self._validators.get(hook_type)
            if validator:
                validator.hooks_by_name = self.hooks_by_name
                
                for hook in self.hooks[hook_type]:
                    await validator.validate(hook)


        



import dill
import time
import asyncio
import statistics
from collections import defaultdict
from typing import Generic, List, Union
from hedra.core.engines.client import Client
from typing_extensions import TypeVarTuple, Unpack
from hedra.core.graphs.events import Event
from hedra.core.engines.types.common.types import RequestTypes
from hedra.core.engines.types.registry import registered_engines
from hedra.core.graphs.hooks.registry.registry_types import (
    ActionHook,
    EventHook, 
    ContextHook,
    TaskHook
)
from hedra.core.graphs.hooks.hook_types.hook_type import HookType
from hedra.core.graphs.hooks.hook_types.internal import Internal
from hedra.core.graphs.stages.types.stage_types import StageTypes
from hedra.core.engines.types.playwright import (
    MercuryPlaywrightClient,
    ContextConfig
)
from hedra.core.personas.persona_registry import get_persona, registered_personas
from hedra.plugins.types.plugin_types import PluginType

from hedra.core.graphs.stages.base.parallel.partition_method import PartitionMethod
from hedra.core.graphs.stages.base.stage import Stage
from .parallel import execute_actions


T = TypeVarTuple('T')


class Execute(Stage, Generic[Unpack[T]]):
    stage_type=StageTypes.EXECUTE

    def __init__(self) -> None:
        super().__init__()
        self.persona = None
        self.client: Client[Unpack[T]] = Client(
            self.graph_name,
            self.graph_id,
            self.name,
            self.stage_id
        )
        
        self.accepted_hook_types = [ 
            HookType.SETUP, 
            HookType.BEFORE, 
            HookType.ACTION,
            HookType.TASK,
            HookType.AFTER,
            HookType.TEARDOWN,
            HookType.CHECK,
            HookType.CHANNEL, 
            HookType.EVENT,
            HookType.CONTEXT
        ]

        self.concurrent_pool_aware_stages = 0
        self.execution_stage_id = 0
        self.optimized = False
        self.execute_setup_stage = None
        self.requires_shutdown = True
        self.allow_parallel = True

    @Internal()
    async def run(self):

        events: List[Union[EventHook, Event]] = [event for event in self.hooks[HookType.EVENT]]
        pre_events: List[EventHook] = [
            event for event in events if isinstance(event, EventHook) and event.pre
        ]
        
        if len(pre_events) > 0:
            pre_event_names = ", ".join([
                event.shortname for event in pre_events
            ])

            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Executing PRE events - {pre_event_names}')
            await asyncio.wait([
                asyncio.create_task(event.call()) for event in pre_events
            ], timeout=self.stage_timeout)

        config = self.client._config
        persona_type_name = config.persona_type.capitalize()

        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Executing - {config.batch_size} - VUs over {self.workers} threads for {config.total_time_string} using - {persona_type_name} - persona')
        await self.logger.spinner.append_message(f'Stage {self.name} executing - {config.batch_size} - VUs over {self.workers} threads for {config.total_time_string} using - {persona_type_name} - persona')

        engine_plugins = self.plugins_by_type.get(PluginType.ENGINE)
        for plugin in engine_plugins.values():
            registered_engines[plugin.name] = plugin
            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Loaded Engine plugin - {plugin.name}')

        persona_plugins = self.plugins_by_type.get(PluginType.PERSONA)

        for plugin_name, plugin in persona_plugins.items():
            registered_personas[plugin_name] = lambda config: plugin(config)
            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Loaded Persona plugin - {plugin.name}')

        if self.workers > 1:

            source_stage_plugins = defaultdict(list)
            for plugin in self.plugins.values():
                source_stage_plugins[plugin.type].append(plugin.name)

            await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Provisioning execution over - {self.workers} - workers')

            action_hooks: List[ActionHook] = self.hooks[HookType.ACTION]
            hooks = [
                {
                    'timeouts': hook.session.timeouts,
                    'reset_connections': hook.session.pool.reset_connections,
                    'hook_name': hook.name,
                    'hook_shortname': hook.shortname,
                    'hook_type': hook.hook_type,
                    'stage': hook.stage,
                    'weight': hook.metadata.weight,
                    'order': hook.metadata.order,
                    **hook.action.to_serializable()
                } for hook in action_hooks
            ]

            task_hooks: List[TaskHook] = self.hooks[HookType.TASK]
            hooks.extend([
                {
                    'timeouts': hook.session.timeouts,
                    'reset_connections': False,
                    'hook_name': hook.name,
                    'hook_shortname': hook.shortname,
                    'hook_type': hook.hook_type,
                    'stage': hook.stage,
                    'weight': hook.metadata.weight,
                    'order': hook.metadata.order,
                    **hook.action.to_serializable()
                } for hook in task_hooks
            ])

            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Starting execution for - {self.workers} workers')             
    
            results_sets = await self.executor.execute_stage_batch(
                execute_actions,
                [
                    dill.dumps({
                        'graph_name': self.graph_name,
                        'graph_path': self.graph_path,
                        'graph_id': self.graph_id,
                        'source_stage_name': self.name,
                        'source_stage_id': self.stage_id,
                        'source_stage_plugins': source_stage_plugins,
                        'partition_method': PartitionMethod.BATCHES,
                        'workers': self.workers,
                        'worker_id': idx + 1,
                        'config': self.client._config,
                        'hooks': hooks
                    }) for idx in range(self.executor.max_workers)
                ]
            )

            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Completed execution for - {self.workers} workers')            
            
            results = []
            elapsed_times = []
            for result_set in results_sets:
                results.extend(result_set.get('results'))
                elapsed_times.append(result_set.get('total_elapsed'))

            total_results = len(results)
            total_elapsed = statistics.median(elapsed_times)

        else:

            start = time.monotonic()

            persona_config = self.client._config
            persona = get_persona(persona_config)
            persona.setup(self.hooks, self.metadata_string)

            action_and_task_hooks: List[Union[ActionHook, TaskHook]] = [
                *self.hooks[HookType.ACTION],
                *self.hooks[HookType.TASK]
            ]

            for hook in action_and_task_hooks:
                if hook.action.type == RequestTypes.PLAYWRIGHT and isinstance(hook.session, MercuryPlaywrightClient):

                    await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Setting up Playwright Session')

                    await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Playwright Session - {hook.session.session_id} - Browser Type: {persona_config.browser_type}')
                    await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Playwright Session - {hook.session.session_id} - Device Type: {persona_config.device_type}')
                    await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Playwright Session - {hook.session.session_id} - Locale: {persona_config.locale}')
                    await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Playwright Session - {hook.session.session_id} - geolocation: {persona_config.geolocation}')
                    await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Playwright Session - {hook.session.session_id} - Permissions: {persona_config.permissions}')
                    await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Playwright Session - {hook.session.session_id} - Color Scheme: {persona_config.color_scheme}')


                    await hook.session.setup(ContextConfig(
                        browser_type=persona_config.browser_type,
                        device_type=persona_config.device_type,
                        locale=persona_config.locale,
                        geolocation=persona_config.geolocation,
                        permissions=persona_config.permissions,
                        color_scheme=persona_config.color_scheme,
                        options=persona_config.playwright_options
                    ))

            results = await persona.execute()

            elapsed = time.monotonic() - start

            await self.logger.filesystem.aio['hedra.core'].info(
                f'{self.metadata_string} - Execution complete - Time (including addtional setup) took: {round(elapsed, 2)} seconds'
            )  

            total_results = len(results)
            total_elapsed = persona.total_elapsed

        post_events: List[EventHook] = [
            event for event in events if isinstance(event, EventHook) and event.pre is False
        ]

        if len(post_events) > 0:
            post_event_names = ", ".join([
                event.shortname for event in post_events
            ])

            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Executing POST events - {post_event_names}')
            await asyncio.wait([
                asyncio.create_task(event.call()) for event in post_events
            ], timeout=self.stage_timeout)


        await self.logger.filesystem.aio['hedra.core'].info( f'{self.metadata_string} - Completed - {total_results} actions at  {round(total_results/total_elapsed)} actions/second over {round(total_elapsed)} seconds')
        await self.logger.spinner.set_default_message(f'Stage - {self.name} completed {total_results} actions at {round(total_results/total_elapsed)} actions/second over {round(total_elapsed)} seconds')

        context_hooks: List[ContextHook] = self.hooks[HookType.CONTEXT]
        context_hooks: List[ContextHook] = self.hooks[HookType.CONTEXT]
        await asyncio.gather(*[
            asyncio.create_task(context_hook.call(self.context)) for context_hook in context_hooks
        ])
        
        return {
            'results': results,
            'total_results': total_results,
            'total_elapsed': total_elapsed
        }


import asyncio
import networkx
import inspect
import threading
import os
from typing import List, Dict, Union, Any, Tuple, Coroutine
from collections import defaultdict
from hedra.core.graphs.events import Event
from hedra.core.graphs.stages.base.stage import Stage
from hedra.core.graphs.stages.error import Error
from hedra.core.graphs.stages.types.stage_types import StageTypes
from hedra.core.graphs.hooks.registry.registrar import registrar
from hedra.core.graphs.hooks.registry.registry_types import EventHook
from hedra.core.graphs.hooks.registry.registry_types.hook import Hook, HookType
from hedra.core.graphs.stages.base.parallel.batch_executor import BatchExecutor
from hedra.core.graphs.transitions.exceptions.exceptions import IsolatedStageError
from hedra.logging import HedraLogger
from hedra.plugins.types.engine.engine_plugin import EnginePlugin
from hedra.plugins.types.reporter.reporter_plugin import ReporterPlugin
from hedra.plugins.types.plugin_types import PluginType
from .transition import Transition
from .common import (
    invalid_transition
)
from .idle import (
    invalid_idle_transition
)


class TransitionAssembler:

    def __init__(
        self, 
        transition_types, 
        graph_name: str=None,
        graph_path: str=None,
        graph_id: str=None,
        cpus: int=None, 
        worker_id: int=None,
        core_config: Dict[str, Any]={}
    ) -> None:
        self.transition_types: Dict[Tuple[StageTypes, StageTypes], Coroutine] = transition_types
        self.core_config = core_config
        self.graph_name = graph_name
        self.graph_path = graph_path
        self.graph_id = graph_id
        self.generated_stages = {}
        self.transitions = {}
        self.instances_by_type: Dict[str, List[Stage]] = {}
        self.cpus = cpus
        self.worker_id = worker_id
        self.loop = asyncio.get_event_loop()
        self.hooks_by_type: Dict[HookType, Dict[str, Hook]] = defaultdict(dict)

        self.logging = HedraLogger()
        self.logging.initialize()

        self._thread_id = threading.current_thread().ident
        self._process_id = os.getpid()

        self._graph_metadata_log_string = f'Graph - {self.graph_name}:{self.graph_id} - thread:{self._thread_id} - process:{self._process_id} - '

    def generate_stages(self, stages: Dict[str, Stage]):

        stages_count = len(stages)
        self.logging.hedra.sync.debug(f'{self._graph_metadata_log_string} - Generating - {stages_count} - stages')
        self.logging.filesystem.sync['hedra.core'].debug(f'{self._graph_metadata_log_string} - Generating - {stages_count} - stages')

        self.instances_by_type = {}

        for stage in stages.values():
            self.instances_by_type[stage.stage_type] = []

        stage_types_count = len(self.instances_by_type)
        self.logging.hedra.sync.debug(f'{self._graph_metadata_log_string} - Found - {stage_types_count} - unique stage types')
        self.logging.filesystem.sync['hedra.core'].debug(f'{self._graph_metadata_log_string} - Found - {stage_types_count} - unique stage types')

        self.generated_stages: Dict[str, Stage] = {
            stage_name: stage() for stage_name, stage in stages.items()
        }

        for stage in self.generated_stages.values():
            stage.core_config = self.core_config
            stage.graph_name = self.graph_name
            stage.graph_path = self.graph_path
            stage.graph_id = self.graph_id

            stage.workers = self.cpus
            stage.worker_id = self.worker_id

            for hook_shortname, hook in registrar.reserved[stage.name].items():
                    hook._call = hook._call.__get__(stage, stage.__class__)
                    setattr(stage, hook_shortname, hook._call)

            methods = inspect.getmembers(stage, predicate=inspect.ismethod) 

            for _, method in methods:
                method_name = method.__qualname__
                hook: Hook = registrar.all.get(method_name)
                
                if hook:
                    hook._call = hook._call.__get__(stage, stage.__class__)
                    setattr(stage, hook.shortname, hook._call)

                    if inspect.ismethod(hook.call) is False:
                        hook.call = hook.call.__get__(stage, stage.__class__)
                        setattr(stage, hook.shortname, hook.call)

                    hook.stage = stage.name
                    hook.stage_instance: Stage = stage
                    
                    self.hooks_by_type[hook.hook_type][hook.name] = hook
                    stage.hooks[hook.hook_type].append(hook)
            
            self.instances_by_type[stage.stage_type].append(stage)

        event_hooks: List[EventHook] = list(self.hooks_by_type.get(HookType.EVENT, {}).values())
        for event_hook in event_hooks:
            for target_hook_name in event_hook.names:    
                target_hook = registrar.all.get(target_hook_name)
                
                if target_hook:
                    self.logging.filesystem.sync['hedra.core'].info(
                        f'{self._graph_metadata_log_string} - Appendng Event - {event_hook.name}:{event_hook.hook_id} - to target Stage - {target_hook.stage}:{target_hook.stage_instance.stage_id} Event Hooks'
                    )
                    
                    event = Event(target_hook, event_hook)
                    target_hook.stage_instance.hooks[HookType.EVENT].append(event)
                    registrar.all[event.name] = event
               
    
        self.logging.hedra.sync.debug(f'{self._graph_metadata_log_string} - Successfully generated - {stages_count} - stages')
        self.logging.filesystem.sync['hedra.core'].debug(f'{self._graph_metadata_log_string} - Successfully generated - {stages_count} - stages')

    def build_transitions_graph(self, topological_generations: List[List[str]], graph: networkx.Graph):

        self.logging.hedra.sync.debug(f'Buiding transitions matrix')
        self.logging.filesystem.sync['hedra.core'].debug(f'Buiding transitions matrix')

        transitions: List[List[Transition]] = []
        plugins: Dict[PluginType, Dict[str, Union[EnginePlugin, ReporterPlugin]]] = {
            PluginType.ENGINE: {},
            PluginType.OPTIMIZER: {},
            PluginType.PERSONA: {},
            PluginType.REPORTER: {}
        }

        for isolate_stage_name in networkx.isolates(graph):
            raise IsolatedStageError(
                self.generated_stages.get(isolate_stage_name)
            )
        
        for generation in topological_generations:

            generation_transitions = []

            stage_pool_size = self.cpus

            stages = {
                stage_name: self.generated_stages.get(stage_name) for stage_name in generation
            }
            parallel_stages = []

            no_workers_stages = [
                StageTypes.WAIT, 
                StageTypes.IDLE
            ]

            stages_count = len(stages)
            self.logging.hedra.sync.debug(f'{self._graph_metadata_log_string} - Provisioning workers - {stages_count} - stages')
            self.logging.filesystem.sync['hedra.core'].debug(f'{self._graph_metadata_log_string} - Provisioning workers- {stages_count} - stages')

            for stage in stages.values():

                for plugin_name, plugin in stage.plugins.items():
                    plugins[plugin.type][plugin_name] = plugin

                if stage.allow_parallel is False and stage.stage_type not in no_workers_stages:
                    stage.workers = 1
                    stage_pool_size -= 1

                    self.logging.hedra.sync.debug(f'{self._graph_metadata_log_string} - Stage - {stage.name} - provisioned - {stage.workers} - workers')
                    self.logging.filesystem.sync['hedra.core'].debug(f'{self._graph_metadata_log_string} - Stage - {stage.name} - provisioned - {stage.workers} - workers')

                    stage.executor = BatchExecutor(stage.workers)

                else:
                    parallel_stages.append((
                        stage.name,
                        stage
                    ))
            
            if len(parallel_stages) > 0:
                batch_executor = BatchExecutor(max_workers=stage_pool_size)

                batched_stages: List[Tuple[str, Stage, int]] = batch_executor.partion_stage_batches(parallel_stages)
                
                for _, stage, assigned_workers_count in batched_stages:

                    self.logging.hedra.sync.debug(f'{self._graph_metadata_log_string} - Stage - {stage.name} - provisioned - {assigned_workers_count} - workers')
                    self.logging.filesystem.sync['hedra.core'].debug(f'{self._graph_metadata_log_string} - Stage - {stage.name} - provisioned - {assigned_workers_count} - workers')

                    stage.workers = assigned_workers_count
                    stage.executor = BatchExecutor(max_workers=assigned_workers_count)

                    stages[stage.name] = stage

            for stage in stages.values():

                stage.plugins_by_type = plugins

                neighbors = list(graph.neighbors(stage.name))

                neighbors_count = len(neighbors)
                self.logging.hedra.sync.debug(f'{self._graph_metadata_log_string} - Discovered - {neighbors_count} - neighboring stages for stage - {stage.name}')
                self.logging.filesystem.sync['hedra.core'].debug(f'{self._graph_metadata_log_string} - Discovered - {neighbors_count} - neighboring stages for stage - {stage.name}')
                
                for neighbor in neighbors:
                    neighbor_stage = self.generated_stages.get(neighbor)

                    transition_action = self.transition_types.get((
                        stage.stage_type,
                        neighbor_stage.stage_type
                    ))

                    if transition_action == invalid_transition or transition_action == invalid_idle_transition:
                        invalid_transition_error, _ = self.loop.run_until_complete(
                            transition_action(stage, neighbor_stage)
                        )

                        raise invalid_transition_error

                    self.logging.hedra.sync.debug(f'{self._graph_metadata_log_string} - Created transition from - {stage.name} - to - {neighbor_stage.name}')
                    self.logging.filesystem.sync['hedra.core'].debug(f'{self._graph_metadata_log_string} - Created transition from - {stage.name} - to - {neighbor_stage.name}')

                    transition = Transition(
                        transition_action,
                        stage,
                        neighbor_stage
                    )

                    generation_transitions.append(transition)

            if len(generation_transitions) > 0:
                transitions.append(generation_transitions)

        self.logging.hedra.sync.debug(f'{self._graph_metadata_log_string} - Transition matrix assemmbly complete')
        self.logging.filesystem.sync['hedra.core'].debug(f'{self._graph_metadata_log_string} - Transition matrix assemmbly complete')

        return transitions

    def map_to_setup_stages(self, graph: networkx.DiGraph):

        self.logging.hedra.sync.debug(f'{self._graph_metadata_log_string} - Mapping stages to requisite Setup stages')
        self.logging.filesystem.sync['hedra.core'].debug(f'{self._graph_metadata_log_string} - Mapping stages to requisite Setup stages')
        
        idle_stages = self.instances_by_type.get(StageTypes.IDLE)
        for idle_stage in idle_stages:
            idle_stage.context.stages = {}
            idle_stage.context.visited = []
            idle_stage.context.results = {}
            idle_stage.context.results_stages = []
            idle_stage.context.summaries = {}
            idle_stage.context.paths = {}
            idle_stage.context.path_lengths = {}
            
        idle_stage_name = idle_stage.__class__.__name__

        complete_stage = self.instances_by_type.get(StageTypes.COMPLETE)[0]

        for stage_type in StageTypes:

            idle_stage.context.stages[stage_type] = {}

            for stage in self.instances_by_type.get(stage_type, []):

                stage_name = stage.__class__.__name__

                has_path = networkx.has_path(
                    graph, 
                    idle_stage_name,
                    stage_name
                )

                if has_path:
                    idle_stage.context.stages[stage_type][stage_name] = stage
                    paths = networkx.all_shortest_paths(graph, stage_name, complete_stage.name)
                
                    stage_paths = []
                    for path in paths:
                        stage_paths.extend(path)
                    
                    idle_stage.context.paths[stage_name] = stage_paths

                    path_lengths = networkx.all_pairs_shortest_path_length(graph)

                    stage_path_lengths = {}
                    for path_stage_name, path_lengths_set in path_lengths:

                        del path_lengths_set[path_stage_name]
                        stage_path_lengths[path_stage_name] = path_lengths_set

                    idle_stage.context.path_lengths[stage_name] = stage_path_lengths.get(stage_name)

        for stage in self.generated_stages.values():
            for idle_stage in idle_stages:
                stage.context = idle_stage.context

        self.logging.hedra.sync.debug(f'{self._graph_metadata_log_string} - Mapped stages to requisite Setup stages')
        self.logging.filesystem.sync['hedra.core'].debug(f'{self._graph_metadata_log_string} - Mapped stages to requisite Setup stages')
    
    def create_error_transition(self, error: Exception):

        from_stage = error.from_stage
            
        error_transition = self.transition_types.get((
            from_stage.stage_type,
            StageTypes.ERROR
        ))

        error_stage = Error()
        error_stage.graph_name = self.graph_name
        error_stage.graph_id = self.graph_id
        error_stage.error = error

        return Transition(
            error_transition,
            from_stage,
            error_stage
        )


from .hook_types import (
    action,
    after,
    before,
    channel,
    check,
    context,
    depends,
    event,
    metric,
    restore,
    save,
    setup,
    task,
    teardown,
    validate
)



from __future__ import annotations
import asyncio
from ctypes import Union
import inspect
from typing import Any, Awaitable, Generic, Optional, TypeVar
from hedra.core.engines.client.config import Config
from hedra.core.engines.types.custom.client import MercuryCustomClient as CustomSession
from hedra.core.engines.types.common import Timeouts
from hedra.core.engines.client.store import ActionsStore
from hedra.plugins.types.plugin_types import PluginType
from hedra.plugins.types.common.event import Event
from hedra.plugins.types.common.plugin_hook import PluginHook
from hedra.plugins.types.common.types import PluginHooks
from hedra.plugins.types.common.plugin import Plugin
from hedra.plugins.types.common.registrar import plugin_registrar
from .action import Action
from .result import Result


A = TypeVar('A')
R = TypeVar('R')


class EnginePlugin(Generic[A, R], Plugin):
    action: Action[A] = None
    result: Result[R] = None
    event: Event[Result[R]] = Event
    security_context: Any = None
    initialized: bool = False
    type=PluginType.ENGINE
    name: str=None

    def __init__(self, config: Config) -> None:

        super(
            EnginePlugin,
            self
        ).__init__()

        self.hooks = {}
        self.request_type = self.__class__.__name__
        self.next_name = None
        self.intercept = False
        self.waiter = None
        self.actions: ActionsStore = None
        self.registered = {}
        self.metadata_string: str = None

        self.config = config
        

        self.action_type = A
        self.request_type = R

        methods = inspect.getmembers(self, predicate=inspect.ismethod) 
        for _, method in methods:

                method_name = method.__qualname__
                hook: PluginHook = plugin_registrar.all.get(method_name)
                
                if hook:
                    hook.call = hook.call.__get__(self, self.__class__)
                    setattr(self, hook.shortname, hook.call)

                    self.hooks[hook.hook_type] = hook

 
        self.session = CustomSession[A, R](
            self,
            concurrency=config.batch_size,
            timeouts=Timeouts(
                connect_timeout=config.connect_timeout,
                total_timeout=config.request_timeout
            ),
            reset_connections=config.reset_connections
        )

    def __getattr__(self, attribute_name: str):

        session = object.__getattribute__(self, 'session')

        if hasattr(session, attribute_name):
            return getattr(session, attribute_name)
            
        else:
            # Default behaviour
            return object.__getattribute__(self, attribute_name)

    async def execute(self, *args, **kwargs) -> Awaitable[Result[R]]:
        
        if self.registered.get(self.next_name) is None:
            action: Action[A] = self.action(
                self.next_name,
                *args,
                **kwargs
            )

            action.plugin_type = self.name
            
            await self.session.prepare(action)

            if self.intercept:
                self.actions.store(self.next_name, action, self)
                
                loop = asyncio.get_event_loop()
                self.waiter = loop.create_future()
                await self.waiter

        return await self.session.execute_prepared_request(
            self.session.registered.get(self.next_name)
        )

    async def close(self):
        close_hook = self.hooks.get(PluginHooks.ON_ENGINE_CLOSE)
        await close_hook.call()


    



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
from .bootstrap_pb2 import (
    ServiceRegistrationStatusRequest,
    ServiceRegistrationStatusResponse,
    ServiceRegisterRequest
)


class BootstrapServerStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.RegisterService = channel.unary_unary(
                '/BootstrapServer/RegisterService',
                request_serializer=ServiceRegisterRequest.SerializeToString,
                response_deserializer=ServiceRegistrationStatusResponse.FromString,
                )
        self.GetServiceRegistrationStatus = channel.unary_unary(
                '/BootstrapServer/GetServiceRegistrationStatus',
                request_serializer=ServiceRegistrationStatusRequest.SerializeToString,
                response_deserializer=ServiceRegistrationStatusResponse.FromString,
                )


class BootstrapServerServicer(object):
    """Missing associated documentation comment in .proto file."""

    def RegisterService(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetServiceRegistrationStatus(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_BootstrapServerServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'RegisterService': grpc.unary_unary_rpc_method_handler(
                    servicer.RegisterService,
                    request_deserializer=ServiceRegisterRequest.FromString,
                    response_serializer=ServiceRegistrationStatusResponse.SerializeToString,
            ),
            'GetServiceRegistrationStatus': grpc.unary_unary_rpc_method_handler(
                    servicer.GetServiceRegistrationStatus,
                    request_deserializer=ServiceRegistrationStatusRequest.FromString,
                    response_serializer=ServiceRegistrationStatusResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'BootstrapServer', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class BootstrapServer(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def RegisterService(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/BootstrapServer/RegisterService',
            ServiceRegisterRequest.SerializeToString,
            ServiceRegistrationStatusResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetServiceRegistrationStatus(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/BootstrapServer/GetServiceRegistrationStatus',
            ServiceRegistrationStatusRequest.SerializeToString,
            ServiceRegistrationStatusResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



import grpc
from hedra.runners.leader_services.proto import DistributedServerStub


class RegisteredLeader:

    def __init__(self, leader_address) -> None:
        self.leader_address = leader_address
        self.leader_id = None

        channel = grpc.aio.insecure_channel(target=leader_address)
        service = DistributedServerStub(channel)

        self.client = service




import time
from numpy import float32, float64, int16, int32, int64
from typing import Any
from datetime import datetime


class AWSTimestreamRecord:

    def __init__(self, record_type: str, record_name: str, record_stage: str, group_name: str, field_name: str, value: Any, session_uuid: str) -> None:

        measure_value_type = None

        if isinstance(value, (float, float32, float64)):
            measure_value_type = "DOUBLE"

        elif isinstance(value, str):
            measure_value_type = "VARCHAR"
        
        elif isinstance(value, bool):
            measure_value_type = "BOOLEAN"

        elif isinstance(value, (int, int16, int32, int64)):
            measure_value_type = "BIGINT"

        elif isinstance(value, datetime):
            value = int(value.timestamp())
            measure_value_type = "TIMESTAMP"

        self.record_type = record_type
        self.record_name = record_name
        self.field = field_name

        self.time = str(int(round(time.time() * 1000)))
        self.time_unit = 'MILLISECONDS'
        self.dimensions = [
            {
                "Name": "type",
                "Value": record_type
            },
            {
                "Name": "source",
                "Value": record_name
            },
            {
                "Name": "stage",
                "Value": record_stage
            },
            {
                "Name": "group",
                "Value": group_name
            },
            {
                "Name": "field_name", 
                "Value": field_name
            },
            {
                "Name": "session_uuid",
                "Value": session_uuid
            }
        ]
        self.measure_name = f'{record_name}_{field_name}_{session_uuid}'
        self.measure_value = str(value)
        self.measure_value_type = measure_value_type

    def to_dict(self):
        return  {
            "Time": self.time,
            "TimeUnit": self.time_unit,
            "Dimensions": self.dimensions,
            "MeasureName": self.measure_name,
            "MeasureValue": self.measure_value,
            "MeasureValueType": self.measure_value_type
        }



import json
from typing import Dict, Any, Union
from hedra.core.engines import registered_engines
from hedra.core.graphs.hooks.registry.registrar import registrar
from hedra.core.engines.client.config import Config
from hedra.core.engines.types.playwright import (
    PlaywrightCommand,
    Page,
    Input,
    URL,
    Options,
    MercuryPlaywrightClient
)
from hedra.core.graphs.hooks.registry.registry_types import ActionHook, TaskHook
from hedra.core.personas.types.default_persona import DefaultPersona
from hedra.core.engines.types.common.types import RequestTypes
from hedra.logging import HedraLogger


async def assemble_playwright_command(
    hook: Union[ActionHook, TaskHook],
    hook_action: Dict[str, Any],
    persona: DefaultPersona,
    config: Config,
    metadata_string: str
):
    logger = HedraLogger()
    logger.initialize()

    action_name = hook_action.get('name')
    hook_type_name = RequestTypes.PLAYWRIGHT.capitalize()

    await logger.filesystem.aio['hedra.core'].debug(
        f'{metadata_string} - Assembling {hook_type_name} Hook - {action_name}'
    )

    action_hooks = hook_action.get('hooks', {})
    action_hook_names = action_hooks.get('names')

    before_hook_name = action_hook_names.get('before')
    after_hook_name = action_hook_names.get('after')
    check_hook_names = action_hook_names.get('checks', [])
    listener_names = action_hook_names.get('listeners', [])
    channel_hook_names = action_hook_names.get('channels', [])

    if before_hook_name:
        await logger.filesystem.aio['hedra.core'].debug(
            f'{metadata_string} - {hook_type_name} Hook - {action_name} - Found Before Hook - {before_hook_name}'
        )

    if after_hook_name:
        await logger.filesystem.aio['hedra.core'].debug(
            f'{metadata_string} - {hook_type_name} Hook - {action_name} - Found After Hook - {after_hook_name}'
        )

    for check_hook_name in check_hook_names:
        await logger.filesystem.aio['hedra.core'].debug(
            f'{metadata_string} - {hook_type_name} Hook - {action_name} - Found Check Hook - {check_hook_name}'
        )

    for channel_hook_name in channel_hook_names:
        await logger.filesystem.aio['hedra.core'].debug(
            f'{metadata_string} - {hook_type_name} Hook - {action_name} - Found Channel Hook - {channel_hook_name}'
        )

    for listener_name in listener_names:
        await logger.filesystem.aio['hedra.core'].debug(
            f'{metadata_string} - {hook_type_name} Hook - {action_name} - Found Listener - {listener_name}'
        )

    hook.session: MercuryPlaywrightClient = registered_engines.get(RequestTypes.PLAYWRIGHT)(
        concurrency=persona.batch.size,
        group_size=config.group_size,
        timeouts=hook_action.get('timeouts')
    )

    await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - {hook_type_name} Hook - {action_name} - Connect Timeout: {hook.session.timeouts.connect_timeout}')
    await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - {hook_type_name} Hook - {action_name} - Read Timeout: {hook.session.timeouts.socket_read_timeout}')
    await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - {hook_type_name} Hook - {action_name} - Total Timeout: {hook.session.timeouts.total_timeout}')
    await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - {hook_type_name} Hook - {action_name} - Group Size: {hook.session.pool.group_size}')

    action_metadata = hook_action.get('metadata', {})

    hook.action: PlaywrightCommand = PlaywrightCommand(
        action_name,
        hook_action.get('command'),
        page=Page(
            **hook_action.get('page', {})
        ),
        url=URL(
            **hook_action.get('url', {})
        ),
        input=Input(
            **hook_action.get('input', {})
        ),
        options=Options(
            **hook_action.get('options', {})
        ),
        user=action_metadata.get('user'),
        tags=action_metadata.get('tags', [])
    )

    for field_name, field_value in hook.action.iter_values():
        if field_value is not None:
            await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - {hook_type_name} Hook - {action_name} - {field_name.capitalize()}: {field_value}')

    await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - {hook_type_name} Hook - {action_name} - User: {hook.action.metadata.user}')

    for tag in hook.action.metadata.tags_to_string_list():
        await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - {hook_type_name} Hook - {action_name} - Tag: {tag}')

    hook.action.hooks.notify = action_hooks.get('notify', False)
    hook.action.hooks.listen = action_hooks.get('listen', False)

    if before_hook_name:
        before_hook = registrar.all.get(before_hook_name)
        hook.action.hooks.before = before_hook.call

    if after_hook_name:
        after_hook = registrar.all.get(after_hook_name)
        hook.action.hooks.after = after_hook.call

    hook.action.hooks.checks = []
    hook.action.hooks.checks.extend(check_hook_names)

    hook.action.hooks.channels = []
    for channel_hook_name in channel_hook_names:
        channel = registrar.all.get(channel_hook_name)
        hook.action.hooks.channels.append(channel.call) 

    hook.action.hooks.listeners = listener_names

    await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - {hook_type_name} Hook - {action_name} - Setup complete')

    return hook


from __future__ import annotations
import threading
import os
import uuid
from collections import defaultdict
from typing import Any, Dict, List, Union
from hedra.core.graphs.simple_context import SimpleContext
from hedra.core.graphs.hooks.registry.registry_types.hook import Hook
from hedra.core.graphs.hooks.hook_types.internal import Internal
from hedra.core.graphs.hooks.hook_types.hook_type import HookType
from hedra.core.graphs.stages.types.stage_states import StageStates
from hedra.core.graphs.stages.types.stage_types import StageTypes
from hedra.core.engines.client.time_parser import TimeParser
from hedra.logging import HedraLogger
from hedra.plugins.types.common.plugin import Plugin
from hedra.plugins.types.plugin_types import PluginType
from hedra.core.graphs.stages.base.parallel.batch_executor import BatchExecutor


class Stage:
    stage_type=StageTypes
    dependencies: List[Stage]=[]
    all_dependencies: List[Stage]=[]
    next_context: Any = None
    stage_timeout=None
    plugins: Dict[str, Plugin] = {}

    def __init__(self) -> None:
        self.name = self.__class__.__name__
        self.graph_path: str = None
        self.stage_id = str(uuid.uuid4())

        self.state = StageStates.INITIALIZED
        self.next_stage: str = None
        self.hooks: Dict[HookType, List[Hook]] = defaultdict(list)
        self.accepted_hook_types: List[HookType] = []
        self.workers: int = None
        self.worker_id: int = 1
        self.generation_stage_names = []
        self.generation_id = 1
        self._shutdown_task = None
        self.requires_shutdown = False
        self.allow_parallel = False
        self.executor: BatchExecutor = None
        self.plugins_by_type: Dict[PluginType, Dict[str, Plugin]] = defaultdict(dict)

        self.core_config = {}
        self.context = SimpleContext()

        self.logger: HedraLogger = HedraLogger()
        self.logger.initialize()

        self.graph_name: str = None
        self.graph_id: str = None

        if self.stage_timeout:
            time_parser = TimeParser(self.stage_timeout)

            self.timeout = time_parser.time
        
        else:
            self.timeout = None

        self.internal_hooks = ['run']

    @Internal()
    async def run(self):
        pass

    @property
    def thread_id(self) -> int:
        return threading.current_thread().ident

    @property
    def process_id(self) -> int:
        return os.getpid()

    @property
    def metadata_string(self):
        return f'Graph - {self.graph_name}:{self.graph_id} - thread:{self.thread_id} - process:{self.process_id} - Stage: {self.name}:{self.stage_id} - '



from typing import List, Union, Dict, Any, Type, Callable, Awaitable
from hedra.core.graphs.hooks.hook_types.hook_type import HookType
from .hook import Hook
from .hook_metadata import HookMetadata


class TaskHook(Hook):

    def __init__(
        self, 
        name: str, 
        shortname: str, 
        call: Callable[..., Awaitable[Any]], 
        weight: int=1, 
        order: int=1, 
        metadata: Dict[str, Union[str, int]]={}, 
        checks: List[Callable[..., Awaitable[Any]]]=[],
        notify: List[str]=[],
        listen: List[str]=[]
    ) -> None:
        super().__init__(
            name, 
            shortname, 
            call, 
            hook_type=HookType.TASK
        )

        self.call: Type[self._call] = self._call
        self.session: Any = None
        self.action: Any = None
        self.notiy = notify
        self.listen = listen
        self.checks = checks
        self.is_notifier = len(notify) > 0
        self.is_listener = len(listen) > 0
        self.notifiers: List[str] = []
        self.listeners: List[str] = []
        self.metadata = HookMetadata(
            weight=weight,
            order=order,
            **metadata
        )


import struct

# Structs for packing and unpacking
_STRUCT_HBBBL = struct.Struct(">HBBBL")
_STRUCT_LL = struct.Struct(">LL")
_STRUCT_HL = struct.Struct(">HL")
_STRUCT_LB = struct.Struct(">LB")
_STRUCT_L = struct.Struct(">L")
_STRUCT_H = struct.Struct(">H")
_STRUCT_B = struct.Struct(">B")


import uuid
from typing import Any, Coroutine, Dict, Iterator, Union, List
from .metadata import Metadata
from .hooks import Hooks
from .types import ProtocolMap

class BaseAction:

    __slots__ = ( 
        'action_id'
        'protocols', 
        'name', 
        'is_setup', 
        'metadata', 
        'hooks'
    )

    def __init__(
        self, 
        name: str=None,
        user: str=None, 
        tags: List[Dict[str, str]] = []
    ) -> None:
        self.action_id = str(uuid.uuid4())
        self.protocols = ProtocolMap()
        self.name = name
        self.is_setup = False
        self.metadata = Metadata(user, tags)
        self.hooks: Hooks[BaseAction] = Hooks()


from hedra.core.graphs.stages import (
    Submit
)

from hedra.reporting.types import (
    KafkaConfig
)


class SubmitKafkaResultsStage(Submit):
    config=KafkaConfig(
        host='localhost:9092',
        client_id='results',
        events_topic='events',
        metrics_topic='metrics'
    )


import asyncio
import socket
from hedra.core.engines.types.common.types import RequestTypes
from hedra.core.engines.types.common.protocols.shared.reader import Reader
from hedra.core.engines.types.common.protocols.shared.writer import Writer
from .protocol import UDPProtocol
from hedra.core.engines.types.common.protocols.shared.constants import _DEFAULT_LIMIT


class UDPConnection:

    def __init__(self, factory_type: RequestTypes = RequestTypes.HTTP) -> None:
        self.loop: asyncio.AbstractEventLoop = asyncio.get_event_loop()
        self.transport = None
        self.factory_type = factory_type
        self._connection = None
        self.socket: socket.socket = None
        self._writer = None

    async def create(self, socket_config=None, *, limit=_DEFAULT_LIMIT, tls=None):

        family, type_, _, _, address = socket_config

        self.socket = socket.socket(family=family, type=type_)
        
        await self.loop.run_in_executor(None, self.socket.connect, address)

        self.socket.setblocking(False)

        reader = Reader(limit=limit, loop=self.loop)
        reader_protocol = UDPProtocol(reader, loop=self.loop)

        self.transport, _ = await self.loop.create_datagram_endpoint(
            lambda: reader_protocol, 
            sock=self.socket
        )

        # TODO: Enable DTLS - This appears to be a *significant* amount of work
        # but would allow for VOIP testing, etc.

        self._writer = Writer(self.transport, reader_protocol, reader, self.loop)
        
        return reader, self._writer


import uuid
from collections import defaultdict
from typing import List
from hedra.logging import HedraLogger
from hedra.reporting.events.types.base_event import BaseEvent
from hedra.reporting.metric import MetricsSet


try:
    from motor.motor_asyncio import AsyncIOMotorClient
    from .mongodb_config import MongoDBConfig
    has_connector = True

except Exception:
    AsyncIOMotorClient = None
    MongoDBConfig = None
    has_connector = False



class MongoDB:

    def __init__(self, config: MongoDBConfig) -> None:
        self.host = config.host
        self.username = config.username
        self.password = config.password
        self.database_name = config.database
        self.events_collection = config.events_collection
        self.metrics_collection = config.metrics_collection
        self.shared_metrics_collection = f'{self.metrics_collection}_common'
        self.errors_collection = f'{self.metrics_collection}_errors'

        self.connection: AsyncIOMotorClient = None
        self.database = None

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = HedraLogger()
        self.logger.initialize()

    async def connect(self):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connecting to MongoDB instance at - {self.host} - Database: {self.database_name}')

        if self.username and self.password:
            connection_string = f'mongodb://{self.username}:{self.password}@{self.host}/{self.database_name}'
        
        else:
            connection_string = f'mongodb://{self.host}/{self.database_name}'

        self.connection = AsyncIOMotorClient(connection_string)
        self.database = self.connection[self.database_name]

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connected to MongoDB instance at - {self.host} - Database: {self.database_name}')

    async def submit_events(self, events: List[BaseEvent]): 
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Events to Collection - {self.events_collection}')
        await self.database[self.events_collection].insert_many(
            [event.record for event in events]
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Events to Collection - {self.events_collection}')

    async def submit_common(self, metrics_sets: List[MetricsSet]):
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Shared Metrics to Bucket - {self.shared_metrics_collection}')
        await self.database[self.shared_metrics_collection].insert_many([
            {
                'name': metrics_set.name,
                'stage': metrics_set.stage,
                'group': 'common',
                **metrics_set.common_stats
            } for metrics_set in metrics_sets
        ])

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Shared Metrics to Bucket - {self.shared_metrics_collection}')

    async def submit_metrics(self, metrics: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Metrics to Bucket - {self.metrics_collection}')

        records = []
        for metrics_set in metrics:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')
            
            for group_name, group in metrics_set.groups.items():
                records.append({
                    'group': group_name,
                    **group.record
                })

        await self.database[self.metrics_collection].insert_many(records)

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Metrics to Bucket - {self.metrics_collection}')

    async def submit_custom(self, metrics_sets: List[MetricsSet]):

        records = defaultdict(list)
        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')
            for custom_group_name, group in metrics_set.custom_metrics.items():
                records[custom_group_name].append({
                    'name': metrics_set.name,
                    'stage': metrics_set.stage,
                    'group': custom_group_name,
                    **group
                })

        for group_name, records in records.items():
            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Custom Metrics to Bucket - {group_name}')

            metrics_collection_name = f'{group_name}_metrics'
            await self.database[metrics_collection_name].insert_many(records)

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Custom Metrics to Bucket - {group_name}')

    async def submit_errors(self, metrics: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Error Metrics to Bucket - {self.errors_collection}')

        for metrics_set in metrics:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            await self.database[self.errors_collection].insert_many([
                {
                    'name': metrics_set.name,
                    'stage': metrics_set.stage,
                    'error_message': error.get('message'),
                    'error_count': error.get('count')
                } for error in metrics_set.errors
            ])

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Error Metrics to Bucket - {self.errors_collection}')

    async def close(self):
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Closing session - {self.session_uuid}')


class ChangedSetting:

    __slots__ = (
        'setting',
        'original_value',
        'new_value'
    )

    def __init__(self, setting, original_value, new_value):
        #: The setting code given. Either one of :class:`SettingCodes
        #: <h2.settings.SettingCodes>` or ``int``
        #:
        #: .. versionchanged:: 2.6.0
        self.setting = setting

        #: The original value before being changed.
        self.original_value = original_value

        #: The new value after being changed.
        self.new_value = new_value

    def __repr__(self):
        return (
            "ChangedSetting(setting=%s, original_value=%s, "
            "new_value=%s)"
        ) % (
            self.setting,
            self.original_value,
            self.new_value
        )



from typing import Dict, List, Tuple, Union
from scipy.optimize import differential_evolution
from .base_algorithm import BaseAlgorithm


class DifferentialEvolutionOptimizer(BaseAlgorithm):

    def __init__(
        self, 
        config: Dict[str, Union[List[Tuple[Union[int, float]]], int]]
    ) -> None:
        super().__init__(config)

    def optimize(self, func):
        return differential_evolution(
            func,
            self.bounds,
            maxiter=self.max_iter
        )




from hedra.runners.leader_services.proto import DistributedServerStub
from .registered_leader import RegisteredLeader


class LeaderRegistry:

    def __init__(self, leader_addresses) -> None:
        logger = Logger()
        self.session_logger = logger.generate_logger()
        self.leader_addresses = leader_addresses
        self.registered_leaders = {}
        self.failed = []
        self.initial_count = len(leader_addresses)
        self.count = 0

    def __getitem__(self, leader_address):
        return self.registered_leaders[leader_address]

    def __iter__(self):
        for leader in self.registered_leaders:
            yield self.registered_leaders[leader]

    async def __aiter__(self):
        for leader in self.registered_leaders:
            yield self.registered_leaders[leader]

    async def initialize(self):
        for leader_address in self.leader_addresses:
            await self.register_leader(leader_address)

    async def register_leader(self, leader_address):
        if leader_address not in self.registered_leaders:
            self.session_logger.info(f'Registering leader from - {leader_address}.')
            self.registered_leaders[leader_address] = RegisteredLeader(leader_address)
            self.count += 1

        if leader_address in self.failed:
            self.session_logger.info(f'Leader at - {leader_address} - re-registered.')
            self.failed.remove(leader_address)

        if leader_address not in self.leader_addresses:
            self.leader_addresses.append(leader_address)

    async def leader_id_not_set(self, leader_address):
        return self.registered_leaders[leader_address].leader_id is None

    async def set_leader_id(self, leader_address, leader_id):
        self.registered_leaders[leader_address].leader_id = leader_id

    async def remove_leader(self, leader_address):
        
        if leader_address in self.registered_leaders and leader_address in self.leader_addresses:
            self.failed.append(leader_address)
            self.leader_addresses.remove(leader_address)
            self.count -= 1

    async def registered_count_equals_expected_count(self):
        return len(self.registered_leaders) == len(self.leader_addresses)

                
        


import inspect
import os
import json
from pathlib import Path
from typing import Optional
from hedra.projects.generation import GraphGenerator
from hedra.core.graphs.stages.base.stage import Stage
from hedra.cli.exceptions.graph.create import InvalidStageType
from hedra.logging import HedraLogger
from hedra.logging import (
    HedraLogger,
    LoggerTypes,
    logging_manager
)


def create_graph(
    path: str, 
    stages: Optional[str], 
    engine: str,
    reporter: str,
    log_level: str
):

    logging_manager.disable(
        LoggerTypes.HEDRA, 
        LoggerTypes.DISTRIBUTED,
        LoggerTypes.FILESYSTEM,
        LoggerTypes.DISTRIBUTED_FILESYSTEM
    )
    logging_manager.update_log_level(log_level)

    logger = HedraLogger()
    logger.initialize()
    logging_manager.logfiles_directory = os.getcwd()

    logger['console'].sync.info(f'Creating new graph at - {path}.')

    if stages is None:
        stages_list = [
            'setup',
            'execute',
            'analyze',
            'submit'
        ]

    else:
        stages_list = stages.split(',')
    
    generated_stages_count = len(stages_list)
    generated_stages = ''.join([
        f'\n-{stage}' for stage in stages_list
    ])   

    logger['console'].sync.info(f'Generating - {generated_stages_count} stages:{generated_stages}') 


    generator = GraphGenerator()

    for stage in stages_list:
        if stage not in generator.valid_types:
            raise InvalidStageType(stage, [
                generator_type_name for generator_type_name, generator_type in generator.generator_types.items() if inspect.isclass(
                    generator_type
                ) and issubclass(generator_type, Stage)
            ])


    with open(path, 'w') as generated_test:
        generated_test.write(
            generator.generate_graph(
                stages_list,
                engine=engine,
                reporter=reporter
            )
        )

    graph_name = path
    if os.path.isfile(graph_name):
        graph_name = Path(graph_name).stem

    hedra_config_filepath = os.path.join(
        os.getcwd(),
        '.hedra.json'
    )

    hedra_config = {}
    if os.path.exists(hedra_config_filepath):
        with open(hedra_config_filepath, 'r') as hedra_config_file:
            hedra_config = json.load(hedra_config_file)

    hedra_graphs = hedra_config.get('graphs', {})

    if hedra_graphs.get(graph_name) is None:
        hedra_graphs[graph_name] = str(Path(path).absolute().resolve())
        with open(hedra_config_filepath, 'w') as hedra_config_file:
            hedra_config['graphs'] = hedra_graphs
            json.dump(hedra_config, hedra_config_file, indent=4)   

    logger['console'].sync.info('\nGraph generated!\n')



from typing import Coroutine
from typing import Any

class AsyncBase:
    def __init__(self, file, loop, executor):
        self._file = file
        self._loop = loop
        self._executor = executor

    async def __aiter__(self):
        """We are our own iterator."""
        return self

    async def __anext__(self):
        """Simulate normal file iteration."""
        line = await self.readline()
        if line:
            return line
        else:
            raise StopAsyncIteration


class AsyncIndirectBase(AsyncBase):
    def __init__(self, name, loop, executor, indirect):
        self._indirect = indirect
        self._name = name
        super().__init__(None, loop, executor)

    @property
    def _file(self):
        return self._indirect()

    @_file.setter
    def _file(self, v):
        pass  # discard writes



from hedra.core.engines.types.http.connection import HTTPConnection


class WebsocketConnection(HTTPConnection):


    def __init__(self, reset_connection: bool = False) -> None:
        super().__init__(reset_connection)

    def iter_headers(self):
        return self.reader.iter_headers()


import asyncio
import os
import inspect
import uvloop
import json
uvloop.install()
import sys
import importlib
import ntpath
from pathlib import Path
from hedra.core.graphs.stages.base.stage import Stage
from hedra.core.graphs import Graph
from hedra.logging import (
    HedraLogger,
    LoggerTypes,
    logging_manager
)


def run_graph(
    path: str, 
    cpus: int, 
    log_level: str, 
    logfiles_directory: str,
    bypass_connection_validation: bool,
    connection_validation_retries: int,
):

    if logfiles_directory is None:
        logfiles_directory = os.getcwd()

    logging_manager.disable(
        LoggerTypes.DISTRIBUTED,
        LoggerTypes.DISTRIBUTED_FILESYSTEM
    )

    logging_manager.update_log_level(log_level)
    logging_manager.logfiles_directory = logfiles_directory

    if os.path.exists(logfiles_directory) is False:
        os.mkdir(logfiles_directory)

    elif os.path.isdir(logfiles_directory) is False:
        os.remove(logfiles_directory)
        os.mkdir(logfiles_directory)

    logger = HedraLogger()
    logger.initialize()

    graph_name = path
    if os.path.isfile(graph_name):
        graph_name = Path(graph_name).stem

    logger['console'].sync.info(f'Loading graph - {graph_name.capitalize()}\n')

    hedra_config_filepath = os.path.join(
        os.getcwd(),
        '.hedra.json'
    )

    hedra_config = {}
    if os.path.exists(hedra_config_filepath):
        with open(hedra_config_filepath, 'r') as hedra_config_file:
            hedra_config = json.load(hedra_config_file)

    hedra_graphs = hedra_config.get('graphs', {})
    hedra_core_config = hedra_config.get('core', {
        'connection_validation_retries': 3
    })

    hedra_core_config['bypass_connection_validation'] = bypass_connection_validation

    if connection_validation_retries:
        hedra_core_config['connection_validation_retries'] = connection_validation_retries
    
    if path in hedra_graphs:
        path = hedra_graphs.get(path)
    
    package_dir = Path(path).resolve().parent
    package_dir_path = str(package_dir)
    package_dir_module = package_dir_path.split('/')[-1]
    
    package = ntpath.basename(path)
    package_slug = package.split('.')[0]
    spec = importlib.util.spec_from_file_location(f'{package_dir_module}.{package_slug}', path)
    
    if path not in sys.path:
        sys.path.append(str(package_dir.parent))

    module = importlib.util.module_from_spec(spec)

    sys.modules[module.__name__] = module

    spec.loader.exec_module(module)
    
    direct_decendants = list({cls.__name__: cls for cls in Stage.__subclasses__()}.values())

    discovered = {}
    for name, stage_candidate in inspect.getmembers(module):
        if inspect.isclass(stage_candidate) and issubclass(stage_candidate, Stage) and stage_candidate not in direct_decendants:
            discovered[name] = stage_candidate

    if hedra_graphs.get(graph_name) is None:
        hedra_graphs[graph_name] = module.__file__


    hedra_config['logging'] = {
        'logfiles_directory': logfiles_directory,
        'log_level': log_level
    }    
    with open(hedra_config_filepath, 'w') as hedra_config_file:
        hedra_config['graphs'] = hedra_graphs
        json.dump(hedra_config, hedra_config_file, indent=4)   

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    logger.filesystem.sync.create_logfile('hedra.core.log')
    logger.filesystem.create_filelogger('hedra.core.log')

    graph = Graph(
        graph_name,
        list(discovered.values()),
        config={
            **hedra_core_config,
            'graph_path': path,
            'graph_module': module.__name__
        },
        cpus=cpus
    )

    graph.assemble()

    loop.run_until_complete(graph.run())

    logger.filesystem.sync['hedra.core'].info(f'{graph.metadata_string} - Completed - {graph.logger.spinner.display.total_timer.elapsed_message}\n')
    logger.console.sync.info(f'\nGraph - {graph_name.capitalize()} - completed! {graph.logger.spinner.display.total_timer.elapsed_message}\n')

    os._exit(0)


import uuid
import time
import asyncio
import psutil
import math
from asyncio import Task
from hedra.core.engines.client.config import Config
from hedra.core.personas.types.default_persona.default_persona import DefaultPersona
from hedra.core.personas.types.types import PersonaTypes


async def cancel_pending(pend: Task):
    try:
        pend.cancel()
        await pend
    
    except asyncio.CancelledError:
        pass


class ConstantSpawnPersona(DefaultPersona):

    def __init__(self, config: Config):
        super(ConstantSpawnPersona, self).__init__(config)

        self.persona_id = str(uuid.uuid4())
        self.action_interval = config.action_interval
        self.type = PersonaTypes.CONSTANT_SPAWN

    async def generator(self, total_time):
        elapsed = 0
        idx = 0
        action_idx = 0
        max_pool_size = math.ceil(self.batch.size * (psutil.cpu_count(logical=False) * 2)/self.workers)

        start = time.time()
        while elapsed < total_time:
            yield action_idx
            
            await asyncio.sleep(self.action_interval)
            elapsed = time.time() - start
            idx += 1

            if idx%self.batch.size == 0:
                await asyncio.sleep(self.batch.interval)

            action_idx = (action_idx + 1)%self.actions_count
            if self._hooks[action_idx].session.active > max_pool_size:
                    try:
                        max_wait = total_time - elapsed
                        await asyncio.wait_for(
                            self._hooks[action_idx].session.wait_for_active_threshold(),
                            timeout=max_wait
                        )
                    except asyncio.TimeoutError:
                        pass



import ssl


def get_http2_ssl_context():
    """
    This function creates an SSLContext object that is suitably configured for
    HTTP/2. If you're working with Python TLS directly, you'll want to do the
    exact same setup as this function does.
    """
    # Get the basic context from the standard library.
    ctx = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)

    # RFC 7540 Section 9.2: Implementations of HTTP/2 MUST use TLS version 1.2
    # or higher. Disable TLS 1.1 and lower.
    ctx.options |= (
        ssl.OP_NO_SSLv2 | ssl.OP_NO_SSLv3 | ssl.OP_NO_TLSv1 | ssl.OP_NO_TLSv1_1
    )

    # RFC 7540 Section 9.2.1: A deployment of HTTP/2 over TLS 1.2 MUST disable
    # compression.
    ctx.options |= ssl.OP_NO_COMPRESSION

    # RFC 7540 Section 9.2.2: "deployments of HTTP/2 that use TLS 1.2 MUST
    # support TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256". In practice, the
    # blocklist defined in this section allows only the AES GCM and ChaCha20
    # cipher suites with ephemeral key negotiation.
    ctx.set_ciphers("ECDHE+AESGCM:ECDHE+CHACHA20:DHE+AESGCM:DHE+CHACHA20")

    # We want to negotiate using NPN and ALPN. ALPN is mandatory, but NPN may
    # be absent, so allow that. This setup allows for negotiation of HTTP/1.1.
    ctx.set_alpn_protocols(["h2", "http/1.1"])

    try:
        if hasattr(ctx, '_set_npn_protocols'):
            ctx.set_npn_protocols(["h2", "http/1.1"])
    except NotImplementedError:
        pass


    ctx.check_hostname = False
    ctx.verify_mode = ssl.CERT_NONE

    return ctx

def get_graphql_ssl_context():
    # Get the basic context from the standard library.
    ctx = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)

    # RFC 7540 Section 9.2: Implementations of HTTP/2 MUST use TLS version 1.2
    # or higher. Disable TLS 1.1 and lower.
    ctx.options |= (
        ssl.OP_NO_SSLv2 | ssl.OP_NO_SSLv3 | ssl.OP_NO_TLSv1 | ssl.OP_NO_TLSv1_1
    )

    # RFC 7540 Section 9.2.1: A deployment of HTTP/2 over TLS 1.2 MUST disable
    # compression.
    ctx.options |= ssl.OP_NO_COMPRESSION

    ctx.minimum_version = ssl.PROTOCOL_TLSv1_2

    # # RFC 7540 Section 9.2.2: "deployments of HTTP/2 that use TLS 1.2 MUST
    # # support TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256". In practice, the
    # # blocklist defined in this section allows only the AES GCM and ChaCha20
    # # cipher suites with ephemeral key negotiation.
    ctx.set_ciphers("ECDHE+AESGCM:ECDHE+CHACHA20:DHE+AESGCM:DHE+CHACHA20")

    # # # We want to negotiate using NPN and ALPN. ALPN is mandatory, but NPN may
    # # # be absent, so allow that. This setup allows for negotiation of HTTP/1.1.
    ctx.set_alpn_protocols(["h2", "http/1.1"])

    try:
        if hasattr(ctx, '_set_npn_protocols'):
            ctx.set_npn_protocols(["h2", "http/1.1"])
    except NotImplementedError:
        pass

    ctx.check_hostname = False
    ctx.verify_mode = ssl.CERT_NONE

    return ctx



def get_default_ssl_context():
    ctx = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
    ctx.check_hostname = False
    ctx.verify_mode = ssl.CERT_NONE

    return ctx
